//**************************************//
//*  Parallel-Serial Converter         *//
//*    (c) AIST & Tohoku Univ.         *//
//*                                    *//
//*    File name   : PARA_SERI.v       *//
//*    Version     : Version 1.0.1     *//
//*    Created     : APRIL/01/2007     *//
//*    Last update : AUGUST/17/2007    *//
//**************************************//

module para_seri (CLK, RSTn_HW, BaudRate, Svld, Dout, Sout,
                  PSrst, SPstatus, PSstatus,Cmd,Status);

input CLK, RSTn_HW, BaudRate, Svld, PSrst, PSstatus;
input [127:0] Dout;
input [3:0]   Cmd,Status;
output Sout, SPstatus;

      //BaudRate = TX_EN(top)
      //Dvld = RD of seri_para

reg   [127:0] Parallel;
reg   [135:0] Parallel_out;
reg   Sout;
reg   LOAD;
reg   [3:0] Count_bit;
reg   [7:0] Seri_bin8;
reg   [4:0] Count16;
reg   [1:0] Send_mode;
reg   SPstatus;

always @(posedge CLK or negedge RSTn_HW) begin
   if (!RSTn_HW) begin
       Sout         <= 1'b1;
       Count_bit    <= 4'h0;
       LOAD         <= 1'b0;
       Count16      <= 5'b00000;
       Send_mode    <= 2'b00;
       Parallel_out <= 136'b0000000000000000000000000000000000;
       SPstatus     <= 1'b0;
   end
   else if (PSrst) begin
       SPstatus <= 1'b0;
   end
   else if (PSstatus) begin
       if (Count16 >= 5'b10001) begin
           LOAD      <= 1'b0;
           Count16   <= 5'b00000;
           SPstatus  <= 1'b1;
           Send_mode <= 2'b00;
       end
       else if (Cmd == 0 && !LOAD) begin
           LOAD <= 1'b1;
       end
       else if (Svld && !LOAD) begin // <-----End signal of seri_para
           Parallel <= Dout;
           LOAD     <= 1'b1;
       end
       else if (BaudRate && LOAD) begin
           if (Send_mode == 2'b00) begin
               Parallel_out <= {Cmd[3:0], Status[3:0], Parallel[127:0]};
               if (Cmd == 4'b0 || Cmd[3] == 1 || Cmd[2] == 1 || Status[3] == 1) begin
                   Parallel_out <= {Cmd[3:0], Status[3:0], 128'b00000000000000000000000000000000};
               end
               Send_mode <= 2'b01;
           end
           else if (Send_mode == 2'b01) begin
               if (Count16 < 5'b10001) begin
                   Seri_bin8    <= Parallel_out[135:128];
                   Parallel_out <= {Parallel_out[127:0], 8'b11111111};
                   Send_mode    <= 2'b10;
               end
           end
           else if(Send_mode == 2'b10) begin
               case(Count_bit)

               0: begin
                   Sout      <= 1'b0;
                   Count_bit <= Count_bit + 4'h1;
               end

               1,2,3,4,5,6,7,8: begin
                   Sout      <= Seri_bin8[0];
                   Seri_bin8 <= {1'b1, Seri_bin8[7:1]};
                   Count_bit <= Count_bit + 4'h1;
               end

               9: begin
                   Sout      <= 1'b1;
                   Count_bit <= 4'h0;  
                   Send_mode <= 2'b01;
                   Count16   <= Count16 +  5'b00001;
               end

               default: begin
                   Sout <= 1'b1;
                   Count_bit <= 4'h0;  
               end

               endcase 
           end
       end
   end
end

endmodule
