//**************************************//
//*  Serial-Parallel Converter         *//
//*    (c) AIST & Tohoku Univ.         *//
//*                                    *//
//*    File name   : SERI_PARA.v       *//
//*    Version     : Version 1.0       *//
//*    Created     : APRIL/01/2007     *//
//*    Last update : AUGUST/16/2007     *//
//**************************************//
`timescale 1ns / 1ps

module seri_para (CLK, RSTn, RSTn_HW, BaudRate16,
                  Sin, Din, Krdy, Drdy, EncDec, EN, SPrst, PSstatus, Cmd);

input CLK, RSTn_HW, BaudRate16, Sin, SPrst;    //BaudRate16=RX_EN(top) *16
output [127:0] Din;
output Krdy, Drdy, EncDec, EN, PSstatus, RSTn;
output [3:0]   Cmd;

reg [135:0] Para_data;
reg [127:0] Din;
reg [7:0]   Parallel;
reg [7:0]   Count_bit;
reg [7:0]   Header;
reg [3:0]   Cmd;
reg [4:0]   Count16;
reg Start_detect, Serial_data;
reg EncDec, Kr, Dr, EN, RSTn;
reg PSstatus;
reg SPstatus;

always @(posedge CLK or negedge RSTn_HW) begin
    if (!RSTn_HW) begin
        Para_data    <= 136'h0000000000000000000000000000000000;
        Din          <= 128'h00000000000000000000000000000000;
        Parallel     <= 8'h00;
        Count_bit    <= 8'h00;
        Start_detect <= 1'b0;
        Count16      <= 5'b00000;
        Kr           <= 1'b0;
        Dr           <= 1'b0;
        EncDec       <= 1'b0;
        EN           <= 1'b1;
        Header       <= 8'b00000000;
        PSstatus     <= 1'b0;
        SPstatus     <= 1'b1;
        RSTn         <= 1'b1;
    end
    else if (SPrst) begin
        PSstatus <= 1'b0;
		  SPstatus <= 1'b1;
    end
    else begin
       if (BaudRate16 && SPstatus) begin 
           if (Count16 >= 5'b10001) begin
               Count16 <= 5'b00000;
               PSstatus <= 1'b1;
               SPstatus <= 1'b0;
           end
           else begin
               if (!Start_detect) begin
                   if (!Sin)
                       Start_detect <= 1'b1;
                end
                else begin
                    Serial_data <= Sin;
                    case(Count_bit)
                    6: begin         //detection of Start_detect bit
                        if (Serial_data == 1)
                            Start_detect <= 1'b0;
                        else
                            Count_bit <= Count_bit + 8'h1;
                    end

                    22, 37, 52, 67, 82, 97, 112, 127: begin  //data bit
                        Count_bit <= Count_bit + 8'h1;
                        Parallel <= {Serial_data, Parallel[7:1]};
                    end

                    140: begin
                        if (Count16 <= 5'b10000)
                            Para_data <= { Para_data[127:0],Parallel};
                        Count_bit <= Count_bit + 8'h1;
                    end

                    141: begin
                        if (Count16 == 5'b10000) begin
                            Header    <= Para_data[135:128];
                            Cmd       <= Para_data[135:132];
                            Din       <= Para_data[127:0];
                        end
                        Count_bit <= Count_bit + 8'h1;
                    end

                    142: begin
                        if (Count16 == 5'b10000) begin
                            if (Header[7:6] == 2'b01) begin
                                EncDec <= 1'b0;
                                Kr     <= 1'b1;
                            end
                        end
                        Count_bit <= Count_bit + 8'h1;
                    end

                    143: begin
                        if (Count16 == 5'b10000) begin
                            if (Header[7] == 1)
                                RSTn   <= 1'b0;
                            else if (Header[6] == 1) begin
                                EncDec <= 1'b1;
                                Kr     <= 1'b1;
                            end
                            else if (Header[5] == 1) begin
                                Dr     <= 1'b1;
                                EncDec <= 1'b0;
                            end
                            else if (Header[4] == 1) begin
                                Dr <= 1'b1;
                                EncDec <= 1'b1;
                            end
                        end
                        Count_bit <= Count_bit + 8'h1;
                    end

                    144: begin
                        if (Count16 == 5'b10000) begin
                            if (Header[7] == 1) begin
                                RSTn <= 1'b1;
                                Header[7] <= 1'b0;
                                PSstatus <= 1'b0;
                                EncDec <= 1'b0;
                            end
                            else if (Header[6] == 1)
                                 Kr <= 1'b0;
                            else if (Header[5] == 1)
                                 Dr <= 1'b0;
                            else if (Header[4] == 1)
                                 Dr <= 1'b0;
                        end
                        Count_bit <= Count_bit + 8'h1;
                    end

                    145: begin
                         Count_bit <= 8'h0;
                         Count16 <= Count16 + 5'b00001;
                         Start_detect <= 1'b0;
                    end

                    default:
                        Count_bit <= Count_bit + 8'h1;

                    endcase 
                end
            end
        end
    end
end

	assign Krdy = Kr & BaudRate16;
	assign Drdy = Dr & BaudRate16;

endmodule